import sys
from acquiscan import extract
from acquiscan.sec_client import SECClient
from acquiscan.scan_utils import (
    EXHIBIT_FORM_TYPES,
    SUBSIDIARY_EXHIBIT_PREFIXES,
    create_result,
    filing_url,
)

def get_subsidiary_docs(sec_client, company_info, filing):
    submission_text = sec_client.get_filing_submission_text(
        company_info["cik"], filing["accessionNumberClean"], filing["accessionNumber"]
    )
    if not submission_text:
        return []
    documents = extract.extract_attached_documents(submission_text)
    return [d for d in documents if d["type"].lower().startswith(SUBSIDIARY_EXHIBIT_PREFIXES)]

def process_exhibit_doc(sec_client, company_info, filing, doc, portfolio, silent):
    if not silent:
        print(f"  Processing {doc['type']}: {doc['filename']}", file=sys.stderr)
    content = sec_client.get_filing_document(
        company_info["cik"], filing["accessionNumberClean"], doc["filename"]
    )
    if not content:
        return
    subs = extract.extract_subsidiaries(content)
    if not silent:
        print(f"  {len(subs)} subsidiaries found", file=sys.stderr)
    url = filing_url(company_info["cik"], filing)
    for name in subs:
        key = name.lower()
        existing = portfolio.get(key)
        if not existing or filing["filingDate"] > existing["filing_date"]:
            portfolio[key] = {
                "name": name,
                "form_type": filing["form"],
                "filing_date": filing["filingDate"],
                "url": url,
            }

def run_structured_exhibit_scan(company_name, config, silent=False, from_date=None, to_date=None):
    sec_client = SECClient(config.user_agent)
    if not silent:
        print(f"Searching for company: {company_name}", file=sys.stderr)
    company_info = sec_client.search_company(company_name)
    if not company_info:
        print("Company not found", file=sys.stderr)
        return create_result({"name": company_name, "cik": None}, portfolio=[])
    if not silent:
        print(f"Found: {company_info['name']} (CIK: {company_info['cik']})", file=sys.stderr)
    filings = sec_client.get_company_filings(
        company_info["cik"], EXHIBIT_FORM_TYPES, from_date, to_date
    )
    if not filings:
        print("No filings found", file=sys.stderr)
        return create_result(company_info, portfolio=[])
    if not silent:
        print(f"Found {len(filings)} filing(s)", file=sys.stderr)

    portfolio = {}
    for i, filing in enumerate(filings, 1):
        if not silent:
            print(
                f"[{i}/{len(filings)}] {filing['form']} ({filing['filingDate']})",
                file=sys.stderr,
            )
        subsidiary_docs = get_subsidiary_docs(sec_client, company_info, filing)
        if not subsidiary_docs:
            if not silent:
                print("  No subsidiary exhibits", file=sys.stderr)
            continue
        for doc in subsidiary_docs:
            process_exhibit_doc(sec_client, company_info, filing, doc, portfolio, silent)

    portfolio_list = sorted(portfolio.values(), key=lambda x: x["name"])
    if not silent:
        print(f"Found {len(portfolio_list)} companies", file=sys.stderr)
    return create_result(company_info, portfolio=portfolio_list)



import argparse
import json
import sys
import traceback
from acquiscan.config import Config
from acquiscan.exhibit_scan import run_structured_exhibit_scan
from acquiscan.llm_scan import run_llm_scan

def format_results(results, mode):
    company = results.get("company", "?")
    lines = []
    for entry in results.get("portfolio", []):
        form = entry.get("form_type", "?")
        date = entry.get("filing_date", "?")
        name = entry.get("name", "?")
        lines.append(f"[{company}] [{mode}] [{form}] [{date}] {name}")
    return "\n".join(lines)

def main():
    parser = argparse.ArgumentParser(
        description="AcquiScan - Extract acquisition information from SEC filings",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Modes:
  exhibits  - Extract subsidiaries from EX-21 and EX-8 exhibit documents (default)
  llm       - Use an LLM to extract structured acquisition data from filings

Examples:
  acquiscan "Apple"
  acquiscan "Meta" --mode llm
  acquiscan "Meta" --mode llm --json --output results.json
"""
    )

    parser.add_argument("company", help="Company name to search for")
    parser.add_argument("--mode", "-m", choices=["exhibits", "llm"], default="exhibits", help="Extraction mode (default: exhibits)")
    parser.add_argument("--json", "-j", action="store_true", help="Output as JSON")
    parser.add_argument("--output", "-o", type=str, default=None, help="Output file path (default: stdout)")
    parser.add_argument("--silent", "-s", action="store_true", help="Suppress progress output")
    parser.add_argument("--from-date", type=str, default=None, help="Only include filings on or after this date (YYYY-MM-DD)")
    parser.add_argument("--to-date", type=str, default=None, help="Only include filings on or before this date (YYYY-MM-DD)")
    parser.add_argument("--include-filings", action="store_true", help="Include per-filing breakdown in JSON output (LLM mode only)")
    parser.add_argument("--include-excerpts", action="store_true", help="Include excerpts in JSON output (LLM mode only, implies --include-filings)")

    args = parser.parse_args()

    if args.include_excerpts:
        args.include_filings = True

    config = Config.load()

    if args.mode == "llm" and not config.llm_api_key:
        print(f"Error: llm_api_key not set in {Config.CONFIG_PATH}", file=sys.stderr)
        sys.exit(1)

    try:
        if args.mode == "exhibits":
            results = run_structured_exhibit_scan(args.company, config, silent=args.silent, from_date=args.from_date, to_date=args.to_date)
        else:
            results = run_llm_scan(args.company, config, silent=args.silent, from_date=args.from_date, to_date=args.to_date, include_excerpts=args.include_excerpts, include_filings=args.include_filings)

        if args.json:
            output = json.dumps(results, indent=2)
        else:
            output = format_results(results, args.mode)

        if args.output:
            with open(args.output, 'w', encoding='utf-8') as f:
                f.write(output)
            print(f"Results written to {args.output}", file=sys.stderr)
        else:
            print(output)

    except KeyboardInterrupt:
        print("\nInterrupted", file=sys.stderr)
        sys.exit(130)
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        if not args.silent:
            traceback.print_exc()
        sys.exit(1)



EXHIBIT_FORM_TYPES = ["10-K", "20-F"]
LLM_FORM_TYPES = ["10-K", "10-Q", "8-K", "20-F", "6-K"]
SUBSIDIARY_EXHIBIT_PREFIXES = ("ex-21", "ex-8")

def filing_url(cik, filing):
    return (
        f"https://www.sec.gov/Archives/edgar/data/{cik}"
        f"/{filing['accessionNumberClean']}/{filing['primaryDocument']}"
    )

def create_result(company_info, portfolio=None, filings=None):
    result = {"company": company_info["name"], "cik": company_info["cik"]}
    if portfolio is not None:
        result["portfolio"] = portfolio
    if filings is not None:
        result["filings"] = filings
    return result



import sys
import json
import openai

REQUIRED_KEYS = {"company_name", "description"}

EXTRACTION_PROMPT = """\
Extract companies/entities acquired or owned by {target_company} from these SEC {form_type} excerpts.

For each, return the company name and a description including, if available:
- acquisition date
- full or partial ownership
- strategic rationale
- any rebranding, renaming, or integration into another product
- any subsequent divestitures

Return JSON: [{{"company_name": "...", "description": "..."}}]
Return [] if none found.

Excerpts:
{excerpts}"""

class LLMClient:
    def __init__(self, config):
        if not config.llm_api_key:
            raise ValueError("llm_api_key not set in ~/.acquiscan.yaml")
        self.client = openai.OpenAI(api_key=config.llm_api_key)
        self.model = config.llm_model

    def extract_acquisitions(self, excerpts, target_company, form_type):
        if not excerpts:
            return []

        prompt = EXTRACTION_PROMPT.format(
            target_company=target_company,
            form_type=form_type,
            excerpts="\n\n---\n\n".join(excerpts)
        )

        try:
            response = self.client.chat.completions.create(
                model=self.model,
                messages=[{"role": "user", "content": prompt}],
            )

            content = response.choices[0].message.content
            if not content:
                return []
            content = content.strip()
            if content.startswith("```"):
                lines = content.split("\n")
                content = "\n".join(lines[1:-1]) if len(lines) > 2 else content

            result = json.loads(content)
            if isinstance(result, dict) and "acquisitions" in result:
                result = result["acquisitions"]
            if not isinstance(result, list):
                return []

            return [item for item in result if isinstance(item, dict) and REQUIRED_KEYS <= item.keys()]
        except json.JSONDecodeError as e:
            print(f"  Error parsing LLM response: {e}", file=sys.stderr)
            return []
        except Exception as e:
            print(f"  Error calling LLM API: {e}", file=sys.stderr)
            return []



import sys
import requests

class SECClient:
    def __init__(self, user_agent):
        self.session = requests.Session()
        self.session.headers.update({
            "User-Agent": user_agent,
            "Accept-Encoding": "gzip, deflate",
        })

    def search_company(self, company_name):
        url = "https://www.sec.gov/files/company_tickers.json"
        try:
            response = self.session.get(url)
            response.raise_for_status()
            companies = response.json()
            company_name_lower = company_name.lower()
            for company_data in companies.values():
                if company_name_lower in company_data.get("title", "").lower():
                    return {
                        "cik": str(company_data.get("cik_str", "")).zfill(10),
                        "name": company_data.get("title", ""),
                        "ticker": company_data.get("ticker", "")
                    }
            return None
        except requests.RequestException as e:
            print(f"Error searching for company: {e}", file=sys.stderr)
            return None

    def get_company_filings(self, cik, form_types, from_date=None, to_date=None):
        url = f"https://data.sec.gov/submissions/CIK{cik}.json"
        try:
            response = self.session.get(url)
            response.raise_for_status()
            data = response.json()
            filings = data.get("filings", {}).get("recent", {})
            matches = []
            all_forms = filings.get("form", [])
            filing_dates = filings.get("filingDate", [])
            accession_numbers = filings.get("accessionNumber", [])
            primary_docs = filings.get("primaryDocument", [])
            for i, form_type in enumerate(all_forms):
                if form_type in form_types:
                    date = filing_dates[i]
                    if from_date and date < from_date:
                        continue
                    if to_date and date > to_date:
                        continue
                    matches.append({
                        "form": form_type,
                        "filingDate": date,
                        "accessionNumber": accession_numbers[i],
                        "accessionNumberClean": accession_numbers[i].replace("-", ""),
                        "primaryDocument": primary_docs[i],
                    })
            return matches
        except requests.RequestException as e:
            print(f"Error retrieving filings: {e}", file=sys.stderr)
            return []

    def get_filing_submission_text(self, cik, accession_number_clean, accession_number):
        try:
            url = f"https://www.sec.gov/Archives/edgar/data/{cik}/{accession_number_clean}/{accession_number}.txt"
            response = self.session.get(url)
            response.raise_for_status()
            return response.text
        except requests.RequestException as e:
            print(f"Error retrieving filing submission text: {e}", file=sys.stderr)
            return None

    def get_filing_document(self, cik, accession_number, document_name):
        try:
            url = f"https://www.sec.gov/Archives/edgar/data/{cik}/{accession_number}/{document_name}"
            response = self.session.get(url)
            response.raise_for_status()
            return response.text
        except requests.RequestException as e:
            print(f"Error downloading filing document: {e}", file=sys.stderr)
            return None



import re
from bs4 import BeautifulSoup

EXCERPT_KEYWORDS = [
    "acquisition", "acquire", "acquiring",
    "merge", "purchase", "buy",
    "subsidiary", "subsidiaries",
]

EXHIBIT_SKIP_PATTERNS = re.compile(
    r'^(exhibit21|exhibit\s|list of\s|subsidiaries|\d+$|\*)', re.IGNORECASE
)

def extract_acquisition_excerpts(html_content):
    if not html_content:
        return []
    soup = BeautifulSoup(html_content, "html.parser")
    excerpts = []
    for div in soup.find_all("div"):
        spans = div.find_all("span", recursive=False)
        if not spans:
            continue
        text = " ".join(span.get_text(strip=True) for span in spans).strip()
        if len(text) < 50:
            continue
        text_lower = text.lower()
        if any(keyword in text_lower for keyword in EXCERPT_KEYWORDS):
            excerpts.append(text)
    return excerpts

def extract_attached_documents(submission_text):
    documents = []
    blocks = re.split(r'<DOCUMENT>', submission_text, flags=re.IGNORECASE)

    def extract_tag(block, tag):
        match = re.search(rf'<{tag}>([^\n\r<]+)', block, flags=re.IGNORECASE)
        return match.group(1).strip() if match else None

    for block in blocks[1:]:
        doc = {
            "type": extract_tag(block, "TYPE"),
            "sequence": extract_tag(block, "SEQUENCE"),
            "filename": extract_tag(block, "FILENAME"),
            "description": extract_tag(block, "DESCRIPTION"),
        }
        if doc["type"] and doc["filename"]:
            documents.append(doc)
    return documents

def extract_subsidiaries(document):
    soup = BeautifulSoup(document, "html.parser")
    tables = soup.find_all("table")
    if tables:
        return _extract_from_tables(tables)
    return _extract_from_divs(soup)

def _extract_from_tables(tables):
    subsidiaries = []
    for table in tables:
        first_non_empty_row_skipped = False
        for tr in table.find_all("tr"):
            for cell in tr.find_all("td"):
                text = cell.get_text(separator=" ", strip=True).replace("\xa0", "")
                if not text:
                    continue
                if not first_non_empty_row_skipped:
                    first_non_empty_row_skipped = True
                    break
                subsidiaries.append(text)
                break
    return subsidiaries

def _extract_from_divs(soup):
    subsidiaries = []
    for div in soup.find_all("div", recursive=True):
        if div.find("div"):
            continue
        text = div.get_text(separator=" ", strip=True).replace("\xa0", "")
        if not text or len(text) < 3:
            continue
        if EXHIBIT_SKIP_PATTERNS.match(text):
            continue
        subsidiaries.append(text)
    return subsidiaries



import sys
from concurrent.futures import ThreadPoolExecutor, as_completed
from acquiscan import extract
from acquiscan.llm_client import LLMClient
from acquiscan.sec_client import SECClient
from acquiscan.scan_utils import (
    LLM_FORM_TYPES,
    create_result,
    filing_url
)

def get_filings_with_excerpts(sec_client, company_info, filings, silent):
    cik = company_info["cik"]
    result_filings = []
    for i, filing in enumerate(filings, 1):
        if not silent:
            print(f"[{i}/{len(filings)}] {filing['form']} ({filing['filingDate']})", file=sys.stderr)
        doc_html = sec_client.get_filing_document(
            cik, filing["accessionNumberClean"], filing["primaryDocument"]
        )
        if not doc_html:
            continue
        excerpts = extract.extract_acquisition_excerpts(doc_html)
        if not excerpts:
            if not silent:
                print("  No relevant excerpts", file=sys.stderr)
            continue
        if not silent:
            print(f"  {len(excerpts)} excerpt(s)", file=sys.stderr)
        result_filings.append({
            "form_type": filing["form"],
            "filing_date": filing["filingDate"],
            "url": filing_url(cik, filing),
            "excerpts": excerpts,
        })
    return result_filings

def get_filing_portfolio(llm_client, company_name, filing_data):
    try:
        acquisitions = llm_client.extract_acquisitions(
            filing_data["excerpts"], company_name, filing_data["form_type"]
        )
        portfolio = []
        for acq in (acquisitions or []):
            name = acq.get("company_name", "").strip()
            if name:
                portfolio.append({
                    "name": name,
                    "description": acq.get("description", ""),
                    "form_type": filing_data["form_type"],
                    "filing_date": filing_data["filing_date"],
                    "url": filing_data["url"],
                })
        return portfolio, None
    except Exception as e:
        return [], e

def _run_parallel_extraction(llm_client, company_name, result_filings, silent, workers=5):
    portfolio = {}
    total = len(result_filings)
    with ThreadPoolExecutor(max_workers=workers) as executor:
        futures = {executor.submit(get_filing_portfolio, llm_client, company_name, fd): fd for fd in result_filings}
        for i, future in enumerate(as_completed(futures), 1):
            filing_data = futures[future]
            filing_portfolio, error = future.result()
            label = f"[{i}/{total}] {filing_data['form_type']} ({filing_data['filing_date']})"
            if error:
                print(f"{label}: Error: {error}", file=sys.stderr)
            elif not silent:
                print(label, file=sys.stderr)
            filing_data["portfolio"] = filing_portfolio
            for entry in filing_portfolio:
                key = entry["name"].lower()
                existing = portfolio.get(key)
                if not existing or entry["filing_date"] > existing["filing_date"]:
                    portfolio[key] = entry
    return portfolio

def run_llm_scan(
    company_name, config, silent=False, from_date=None, to_date=None,
    include_excerpts=False, include_filings=False,
):
    sec_client = SECClient(config.user_agent)
    if not silent:
        print(f"Searching for company: {company_name}", file=sys.stderr)
    company_info = sec_client.search_company(company_name)
    if not company_info:
        print("Company not found", file=sys.stderr)
        return create_result({"name": company_name, "cik": None}, portfolio=[])
    if not silent:
        print(f"Found: {company_info['name']} (CIK: {company_info['cik']})", file=sys.stderr)
    filings = sec_client.get_company_filings(
        company_info["cik"], LLM_FORM_TYPES, from_date, to_date
    )
    if not filings:
        print("No filings found", file=sys.stderr)
        return create_result(company_info, portfolio=[])
    if not silent:
        print(f"Found {len(filings)} filing(s)", file=sys.stderr)

    try:
        llm_client = LLMClient(config)
    except ValueError as e:
        print(f"Error: LLM not available: {e}", file=sys.stderr)
        return create_result(company_info, portfolio=[])

    result_filings = get_filings_with_excerpts(sec_client, company_info, filings, silent)
    if not result_filings:
        print("No excerpts found in any filings", file=sys.stderr)
        return create_result(company_info, portfolio=[])

    if not silent:
        print("Extracting acquisitions with LLM...", file=sys.stderr)
    portfolio = _run_parallel_extraction(llm_client, company_info["name"], result_filings, silent)

    if not include_excerpts:
        for fd in result_filings:
            fd.pop("excerpts", None)

    portfolio_list = list(portfolio.values())
    if not silent:
        print(f"Found {len(portfolio_list)} companies", file=sys.stderr)

    result = create_result(company_info, portfolio=portfolio_list)
    if include_filings:
        result["filings"] = result_filings
    return result



import sys
from pathlib import Path
import yaml

CONFIG_PATH = Path.home() / ".acquiscan.yaml"

DEFAULT_CONFIG = """\
# Required: SEC EDGAR requires declaring a user agent for all requests.
# https://www.sec.gov/about/webmaster-frequently-asked-questions#developers
user_agent: "Your Name (your@email.com)"

# Optional: only needed for --mode llm
# llm_api_key: "sk-..."
# llm_model: "gpt-5-mini"
"""

DEFAULTS = {
    "user_agent": None,
    "llm_api_key": None,
    "llm_model": "gpt-5-mini",
}

class Config:
    CONFIG_PATH = CONFIG_PATH

    def __init__(self, **kwargs):
        self.user_agent = kwargs.get("user_agent", DEFAULTS["user_agent"])
        self.llm_api_key = kwargs.get("llm_api_key", DEFAULTS["llm_api_key"])
        self.llm_model = kwargs.get("llm_model", DEFAULTS["llm_model"])

    @classmethod
    def load(cls):
        if not CONFIG_PATH.exists():
            CONFIG_PATH.write_text(DEFAULT_CONFIG, encoding="utf-8")
            print(
                f"Created config file: {CONFIG_PATH}\n"
                f"Please edit it with your details and run again.",
                file=sys.stderr,
            )
            sys.exit(1)

        with open(CONFIG_PATH, encoding="utf-8") as f:
            data = yaml.safe_load(f) or {}

        config = cls(**data)
        if not config.user_agent or "your@email.com" in config.user_agent.lower():
            print(
                f"Please set your user_agent in {CONFIG_PATH}",
                file=sys.stderr,
            )
            sys.exit(1)
        return config
